# Main script for Boundary Layer eq. solver.
# VERSION: 1
# AUTHOR: Vojtech Pezlar
# LAST UPDATE: 04/05/2024
# EMAIL: pezlavoj@student.cvut.cz

#-----------------------------------------------------------------------------------------------------------------------------#
# Code Purpose:                                                                                                               #
# The following code solves the Blasius Boundary Layer (BL) equations, using shooting methods. The solution is marched        #
# from the wall towards specified height, given initial value for wall shear. The value of velocity derivative at eta_max     #
# is checked and if its is not 1 (Merge condition), the value of wall shear is adjusted using Secant root-finding method.     #
#-----------------------------------------------------------------------------------------------------------------------------#


# -----------------------------------------------#
#                    Libraries                   #
# -----------------------------------------------#

# 3rd party libraries
import numpy as np                     # General array and vector manipulation
import scipy as sp                     # Used in the end to calculate BL thickness for comparison with literature
from matplotlib import pyplot as plt   # Plotting and visualization

# -----------------------------------------------#
#                    Freestream                  #
# -----------------------------------------------#
# BL Height (non-dimensional)
eta_final = 7.5

# -----------------------------------------------#
#                 Solver Settings                #
# -----------------------------------------------#
# Secant root finding method requires 2 initial guesses. Guesses used are the standard approximation values
# Initial Guess values
shear00 = 1 / 3
shear01 = 1 / 3 + 0.01


# -----------------------------------------------#
#                    Equations                   #
# -----------------------------------------------#
# BL Eq. System Definition
def blasius(eta):
    return np.array([eta[1], eta[2], -eta[0] * eta[2]])

# *hint -> to convert to compressible system, add more equations as:
# np.array([eq_1, eq_2, eq_3, ..., eq_n])


# -----------------------------------------------#
#                   Integrator                   #
# -----------------------------------------------#
# Runge-Kutta 4 Integrator
def RK4(F, interval, step, S0, condition):

    # number of steps based on step size
    n_steps = int(interval[1] / step)
    interval_vector = np.linspace(interval[0], interval[1], n_steps)

    # Allocating state vector and inputting initial conditions
    q_state = np.zeros((n_steps, 3))
    q_state[0, :] = S0

    # *hint when expanding to compressible case, the state vector (q_state) size changes (not only 3 equations anymore)

    # March for specified number of steps, given by n_steps
    for i in range(n_steps - 1):

        # Obtaining RK4 Coefficients
        F1 = F(q_state[i, :])
        F2 = F(q_state[i, :] + (step / 2) * F1)
        F3 = F(q_state[i, :] + (step / 2) * F2)
        F4 = F(q_state[i, :] + step * F3)

        # next step
        q_state[i + 1, :] = q_state[i, :] + (step / 6) * (F1 + 2 * F2 + 2 * F3 + F4)

    # Decide if only the state is outputted or vector containing the steps is also included.
    # vector containing the steps is used in the end for plotting
    if condition == "True":
        return [q_state, interval_vector]
    else:
        return q_state


# -----------------------------------------------#
#            Shooting Method Solution            #
# -----------------------------------------------#
# Shooting Method Settings
delta = 0.01  # Step size
tol = 1e-6    # Solution tolerance
err = 1       # Initialization of control variable for root-finding method

# Secant technique (modified Newton), Root finding method
while tol < err:
    # Marching at (n-1) and (n) guess values
    q00 = RK4(blasius, [0, eta_final], delta, [0, 0, shear00], 'None')
    q01 = RK4(blasius, [0, eta_final], delta, [0, 0, shear01], 'None')

    # *hint when expanding to compressible case, the state vector size changes (not only 3 equations anymore),
    # also applies to initial conditions

    # Fitness function -> compares the value of the velocity derivative at the final step with desired value of 1
    err = abs(q01[-1, 1] - 1)

    # Secant method formulas
    denominator = ((q01[-1, 1] - 1) - (q00[-1, 1] - 1)) / (shear01 - shear00)
    # modified next value
    next_val = shear01 - (q01[-1, 1] - 1) / denominator

    # Updating values
    shear00 = shear01
    shear01 = next_val


# -----------------------------------------------#
#                Resulting Profile               #
# -----------------------------------------------#
# Obtaining solution for the converged initial condition
q, y = RK4(blasius, [0, eta_final], delta, [0, 0, shear01], 'True')

# # Using integration to obtain boundary layer thickness
# Function to be integrated (Definition of incompressible non-dimensional BL thickness)
thick_func = 1 - q[:, 1]
# Integration
BLTh = sp.integrate.trapezoid(thick_func, x=y, axis=0)


# -----------------------------------------------#
#                Printing Results                #
# -----------------------------------------------#
print("Wall shear   = " + "% .5f" % q[0, 2])
print("BL thickness = " + "% .5f" % BLTh)


# -----------------------------------------------#
#                   Visualization                #
# -----------------------------------------------#
plt.plot(y, q[:, 0])
plt.plot(y, q[:, 1])
plt.plot(y, q[:, 2])
plt.xlabel("eta")
plt.legend(['f(eta)', "f'(eta)", "f''(eta)"])
plt.show()
